use std::marker::PhantomData;

use zerocopy::{AsBytes, FromBytes, FromZeroes, Unaligned};

use crate::ReaderTrait;
use crate::{transaction, FilePointer, TransactionHandle, U64};

#[derive(Clone, Copy, FromBytes, FromZeroes, AsBytes, Unaligned)]
#[repr(packed)]
pub struct Queue<T: FromBytes + FromZeroes + AsBytes + Unaligned + Clone + Copy> {
    pub(crate) head: FilePointer<QueueElement<T>>,
    pub(crate) length: U64,
    _phantom: PhantomData<T>,
}

impl<T: FromBytes + FromZeroes + AsBytes + Unaligned + Clone + Copy> Queue<T> {
    pub fn new<R>(transaction: &mut TransactionHandle<R>) -> FilePointer<Queue<T>> {
        let (queue, data) = transaction.allocate();
        *data = Self {
            head: FilePointer::null(),
            length: 0.into(),
            _phantom: PhantomData,
        };
        queue
    }
}

impl<T: FromBytes + FromZeroes + AsBytes + Unaligned + Clone + Copy> FilePointer<Queue<T>> {
    pub fn enqueue<R>(self, transaction: &mut TransactionHandle<R>, item: T) -> Self {
        self.enqueue_many(transaction, &[item])
    }

    pub fn dequeue<R>(self, transaction: &mut TransactionHandle<R>) -> Option<(Self, T)> {
        let mut res = None;
        self.dequeue_many_inner(transaction, 1, |t| assert!(res.replace(t).is_none()))
            .map(|ptr| (ptr, res.unwrap()))
    }

    pub fn enqueue_many<R>(self, transaction: &mut TransactionHandle<R>, items: &[T]) -> Self {
        if items.is_empty() {
            return self;
        }

        let old = *transaction.read(self);

        let mut head = old.head;

        for item in items {
            let (element, data) = transaction.allocate();
            *data = QueueElement {
                next: head,
                data: *item,
                _phantom: PhantomData,
            };

            head = element;
        }

        let (queue, data) = transaction.modify(self);
        *data = Queue {
            head,
            length: (old
                .length
                .get()
                .checked_add(items.len().try_into().unwrap())
                .unwrap())
            .into(),
            _phantom: PhantomData,
        };
        queue
    }

    pub fn dequeue_many<R>(
        self,
        transaction: &mut TransactionHandle<R>,
        n: u64,
    ) -> Option<(Self, Vec<T>)> {
        let mut res = Vec::with_capacity(n.try_into().unwrap());
        self.dequeue_many_inner(transaction, n, |t| res.push(t))
            .map(|ptr| {
                res.reverse();
                (ptr, res)
            })
    }

    // NOTE: calls f with the elements in reverse order.
    fn dequeue_many_inner<R>(
        self,
        transaction: &mut TransactionHandle<R>,
        n: u64,
        mut f: impl FnMut(T),
    ) -> Option<Self> {
        if n == 0 {
            return Some(self);
        }

        let old = *transaction.read(self);

        if old.length.get() < n {
            return None;
        }

        let (queue, data) = transaction.modify(self);

        *data = Queue {
            head: old.head,
            length: (old.length.get().checked_sub(n).unwrap()).into(),
            _phantom: PhantomData,
        };

        // println!("length: {}", old.length.get());

        let mut ptr = data.head;

        let skip_n = old.length.get().checked_sub(n).unwrap();

        // println!("skip_n: {skip_n}");

        // skip the head
        for i in 0..skip_n {
            // println!("skip [{i}] {ptr:?}");

            ptr = *transaction.read(field_ptr!(ptr, QueueElement<T>, next));
        }

        // remove the tail
        for i in 0..n {
            // println!("free [{i}] {ptr:?}");

            let element = *transaction.read(ptr);
            transaction.free(ptr);

            f(element.data);
            ptr = element.next;
        }

        Some(queue)
    }

    pub fn length(self, reader: &impl ReaderTrait) -> u64 {
        reader.read(field_ptr!(self, Queue<T>, length)).get()
    }
}

#[derive(Clone, Copy, FromBytes, FromZeroes, AsBytes, Unaligned)]
#[repr(packed)]
pub(crate) struct QueueElement<T: FromBytes + FromZeroes + AsBytes + Unaligned> {
    pub(crate) next: FilePointer<QueueElement<T>>,
    pub(crate) data: T,
    _phantom: PhantomData<T>,
}
